///////////////////////////////////////////////////////////
// Complex FIR Filter example for Xilinx MicroBlaze FSL
// 
// Created by Bruce Karsten of Xilinx (bruce.karsten@xilinx.com)
//

#include <stdio.h>
#include "ComplexFilter.h"
#include "MB_iDEN_Filter.h"

#if defined(MICROBLAZE)
#include "xparameters.h"
#include "mb_interface.h"
#include "xbasic_types.h"
#include "xutil.h"
#include "xtmrctr.h"
#define printf xil_printf
#define TIMER_FREQ_MILLI (XPAR_CPU_CORE_CLOCK_FREQ_HZ/1000)
#define INIT_TIMER() \
  XTmrCtr_Initialize(&TimerCounter, XPAR_MYTIMER_DEVICE_ID);\
  XTmrCtr_SetResetValue(&TimerCounter,0,0);\
  XTmrCtr_Start(&TimerCounter,0)
#define START_TIMER(var) XTmrCtr_Reset(&TimerCounter,0)
#define END_TIMER(var) var=XTmrCtr_GetValue(&TimerCounter,0)
#define TIMER_UNITS "cycles"
	XTmrCtr		TimerInst;
	Xuint32		StartIF_TS;
	Xuint32		FinishIF_TS;
	Xuint32		t_Time1, t_Time2;
	Xuint32		TimeSA, TimeHA;
	XStatus		status;
#endif

int k;
// Mass Storage
int32 input_data[NUM_SAMPS_IN_SLOT];
int32 if_filter_coef[IF_FILT_LEN];
int32 if_filter_output[NUM_SAMPS_IN_SLOT];

// Software ComplexFIR running in MicroBlaze...
void software_fir(void) {
	int32	coef_mem[IF_FILT_LEN];
	int32	filt_hist[IF_FILT_LEN];
	int32	inSample;
	int32	outFilter;
	int32		i, jj, ii;
	int32		write_idx;
	int32		read_idx;
	int32		t_read_idx, t_read_idx2;
	int32	t_coef, t_hist;
	int16	coef_real, coef_imag;
	int16	hist_real, hist_imag;
	int32	p1_real, p2_real;
	int32	p1_imag, p2_imag;
	int32	res_real, res_imag;
	int32	accum_real, accum_imag;

#if defined(MICROBLAZE)
	printf ("\r\n===================Running the Software-Only Version================\r\n");
	printf ("--> Begin filtering two slots\r\n");
	StartIF_TS = XTmrCtr_GetValue (&TimerInst, 0);
#endif

	// Run NUM_SLOTS slots
	for (ii=0;ii < NUM_SLOTS;ii++){
		// Get StartIF Timestamp
#if defined(MICROBLAZE)
		// reset the timer...
//		XTmrCtr_Reset (&TimerInst, 0);
#else
		printf ("Begin Filtering a Slot");
#endif
	
		// Init history pointers
		write_idx = IF_FILT_LEN - 1;
		read_idx = 0;

		for (i=0;i<IF_FILT_LEN;i++) {
			// Fill the Coefficient Memory
			coef_mem[i] = if_filter_coef[i];
			// Clear the filter history
			filt_hist[i++] = 0;
		}

		// Begin to process incoming data
		for (jj=0;jj<NUM_SAMPS_IN_SLOT;jj++){
#if defined(MICROBLAZE)
			if ((jj & 0xFFFFFF80) == jj) printf("*");
#endif
			filt_hist[write_idx] = input_data[jj];
			if (write_idx == (IF_FILT_LEN - 1))
					write_idx = 0;
				else
					write_idx++;

			t_read_idx = read_idx++;
			if (read_idx >= IF_FILT_LEN )
					read_idx = 0;

			//
			// Create Filtered value
			//
			
			// Clear the accums
			accum_real = 0;
			accum_imag = 0;

			i=0;
			do {
			// Get the pieces
				t_coef = coef_mem[i++];
				t_hist = filt_hist[t_read_idx++];

				if (t_read_idx >= IF_FILT_LEN )
					t_read_idx = 0;

				coef_real = (t_coef >> 16) & 0xFFFF;
				coef_imag = (t_coef & 0xFFFF);
				hist_real = (t_hist >> 16) & 0xFFFF;
				hist_imag = (t_hist & 0xFFFF);

				// Create the products
				p1_real = coef_real * hist_real;
				p1_imag = coef_real * hist_imag;
				p2_real = coef_imag * hist_imag;
				p2_imag = coef_imag + hist_real;

				// Combine products & reduce bits
				res_real = p1_real - p2_real;
				res_imag = p1_imag + p2_imag;
				res_real >>= 12;
				res_imag >>= 12;
				
				// Accumulate products
				accum_real += res_real;
				accum_imag += res_imag;
			} while (i<IF_FILT_LEN);
			// Return result to 16 bits (assuming > 256 length filter)
			accum_real >>= 13;
			accum_imag >>= 13;
			
			// Transfer result to output word
			outFilter = ((accum_real & 0xFFFF) << 16) + (accum_imag & 0xFFFF);
				
			// Output Filtered value
			if_filter_output[jj] = outFilter;
		}
	}

		// Get FinishIF Timestamp
#if defined(MICROBLAZE)
		FinishIF_TS = XTmrCtr_GetValue (&TimerInst, 0);
		printf ("\r\n--> Done filtering two slots");

	// Print Accumulated Stats
	TimeSA = (FinishIF_TS - StartIF_TS)/TIMER_FREQ_MILLI; 
	
	printf (", execution time : %d milliseconds\r\n", TimeSA);

#else
		printf ("Complete Filtering a Slot\r\n");
#endif	
	}

// Impulse declarations
extern co_architecture co_initialize(void *);

int main(void)
{
	int c;

	IF_SIM ( cosim_logwindow swlog; )
	IF_SIM ( swlog = cosim_logwindow_create ("complex_fir"); )
	IF_SIM (cosim_logwindow_fwrite (swlog, 
			"Begin Simulation\n");)
	
	// Initialize input data array
	InitInputData ( &input_data[0], (short) NUM_SAMPS_IN_SLOT);

	// Initialize IF Filter Coefficients
	InitIFFiltCoef ( &if_filter_coef[0], (short) IF_FILT_LEN);
	
#if defined(MICROBLAZE)
	// Enable ICache
	microblaze_init_icache_range(0, XPAR_MICROBLAZE_0_CACHE_BYTE_SIZE);
   microblaze_enable_icache();
    
	// Enable DCache
   microblaze_init_dcache_range(0, XPAR_MICROBLAZE_0_DCACHE_BYTE_SIZE);
   microblaze_enable_dcache();
	
	printf("\r\n\n\n\n");
	printf("    ____                      __             ______\r\n");
	printf("   /  _/___ ___  ____  __  __/ /_______     / ____/   C-toFPGA Tools\r\n");
	printf("   / // __ `__ \\/ __ \\/ / / / / ___/ _ \\   / /        for Xilinx FPGA\r\n");
	printf(" _/ // / / / / / /_/ / /_/ / (__  )  __/  / /___      Platforms\r\n");
	printf("/___/_/ /_/ /_/ .___/\\__,_/_/____/\\___/   \\____/\r\n");
	printf("             /_/\r\n");
	printf("--------------------------------------------------------------------\r\n");
	printf("Complex FIR Filter Acceleration demonstration, featuring the Xilinx\r\n");
	printf("MicroBlaze, Virtex-5 FPGA and Impulse C-to-FPGA tools.\r\n");

	// Initialize Timer
	status = XTmrCtr_Initialize (&TimerInst, XPAR_XPS_TIMER_1_DEVICE_ID);
	if (status != XST_SUCCESS)
		printf ("ERROR Initializing Timer, returned code : %d\r\n", status);
	XTmrCtr_Start (&TimerInst, 0);
	t_Time1 = XTmrCtr_GetValue (&TimerInst, 0);
	t_Time2 = XTmrCtr_GetValue (&TimerInst, 0);
#endif

	// Running on software
	software_fir();

#if defined(MICROBLAZE)
	printf ("------> Visit www.ImpulseC.com to learn more!");

	// Disable DCache
	microblaze_disable_dcache();
    microblaze_init_dcache_range(0, XPAR_MICROBLAZE_0_DCACHE_BYTE_SIZE);
	
	// Disable ICache
	microblaze_disable_icache();
    microblaze_init_icache_range(0, XPAR_MICROBLAZE_0_CACHE_BYTE_SIZE);
	
	return 0;
#else
	printf ("COMPLETE APPLICATION\r\n");
	printf ("Press Enter to continue...\r\n");
	c = getc(stdin);
#endif
}
